# -*- coding: utf-8 -*-
"""
Created on Tue Jan 23 13:28:19 2024

@author: Ju Feng (jufen@dtu.dk)
"""

import numpy as np
from topdesign import Model
from topdesign import rotate_coordinates



class WFModel(Model):
    def __init__(self, wf_obj,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['AEP'], 
                 parameters_dict=None, name='wf_model', info=None):
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.wf_obj = wf_obj



class WFModel_Floris(WFModel):
    """ Wind farm model using the FLORIS model (work with floris 4, 
    floris v4.2.2 has been used in the example run).
    """
    def __init__(self, wf_obj,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['AEP'], 
                 parameters_dict=None, name='wf_model_floris', info=None):
        super().__init__(wf_obj, input_arg_keys, output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
    
        
    def func(self, input_dict):
        layout_x = input_dict['layout_x']
        layout_y = input_dict['layout_y']

        self.wf_obj.set(
            layout_x = layout_x,
            layout_y = layout_y
        )

        # Compute the AEP (GWh) using the default settings
        self.wf_obj.run()
        aep = self.wf_obj.get_farm_AEP() / 1.0e9
        
        output_dict = {'AEP': aep}
        return output_dict



class WFModel_PyWake(WFModel):
    """ Wind farm model using the PyWake model.
    """
    def __init__(self, wf_obj,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['AEP'], 
                 parameters_dict=None, name='wf_model_pywake', info=None):
        super().__init__(wf_obj, input_arg_keys, output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
    
        
    def func(self, input_dict):
        layout_x = input_dict['layout_x']
        layout_y = input_dict['layout_y']

        if self.parameters_dict is None:
            sim_res = self.wf_obj(x=np.array(layout_x), 
                                  y=np.array(layout_y))     # wind turbine positions
        
        else:
            wd = self.parameters_dict['wd']
            ws = self.parameters_dict['ws']
            sim_res = self.wf_obj(x=np.array(layout_x), 
                                  y=np.array(layout_y),
                                  wd=wd, ws=ws) 

        # Compute the AEP (GWh) using the default settings
        aep = sim_res.aep(normalize_probabilities=False)
        aep_tot = aep.values.sum()
        
        output_dict = {'AEP': aep_tot}
        return output_dict



class RectangularBoundary(Model):
    def __init__(self, x_cen, y_cen, width, height, rotation_in_deg=0.0,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['viol_boundary'], 
                 parameters_dict=None, name='rectangular_boundary', info=None,
                 viol_buffer=1e-3):
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.x_cen = x_cen
        self.y_cen = y_cen
        self.width = width
        self.height = height
        self.rotation_in_deg = rotation_in_deg
        self.viol_buffer = viol_buffer


    def cal_viol_degree(self, x, y):
        x_list = np.asarray(x).ravel()
        y_list = np.asarray(y).ravel()
        viol_degree = np.zeros_like(x_list) + 0.0

        x_adjusted, y_adjusted = rotate_coordinates(
            x_list - self.x_cen, y_list - self.y_cen, 
            np.deg2rad(self.rotation_in_deg))

        for i in range(len(x_list)):
            x_adj = x_adjusted[i]
            y_adj = y_adjusted[i]

            dist2x = max(0.0, x_adj - self.width/2.0, -self.width/2.0 - x_adj)
            dist2y = max(0.0, y_adj - self.height/2.0, -self.height/2.0 - y_adj)
            viol_degree[i] = np.sqrt(dist2x**2 + dist2y**2)

            # Do not consider the violation if it is within the buffer range
            if viol_degree[i] > 0 and viol_degree[i] <= self.viol_buffer:
                viol_degree[i] = 0.0
        
        if np.isscalar(x):
            return viol_degree[0]
        elif isinstance(x, np.ndarray):
            return viol_degree.reshape(x.shape)
        else:
            return  [viol_degree[i] for i in range(len(x_list))]

    
    def func(self, input_dict):
        x = input_dict['layout_x']
        y = input_dict['layout_y']

        viol_degree = np.sum(self.cal_viol_degree(x, y))

        output_dict = {'viol_boundary': viol_degree}

        return output_dict


    def generate_boundary_points(self):
        x_boundary = np.array([self.x_cen - self.width/2, 
                               self.x_cen + self.width/2, 
                               self.x_cen + self.width/2, 
                               self.x_cen - self.width/2,
                               self.x_cen - self.width/2])
        y_boundary = np.array([self.y_cen - self.height/2, 
                               self.y_cen - self.height/2, 
                               self.y_cen + self.height/2, 
                               self.y_cen + self.height/2,
                               self.y_cen - self.height/2])
        x_boundary_rot, y_boundary_rot = rotate_coordinates(
            x_boundary, y_boundary, 
            -np.deg2rad(self.rotation_in_deg))
        
        return x_boundary_rot, y_boundary_rot

    


class CircularBoundary(Model):
    def __init__(self, x_cen, y_cen, radius,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['viol_boundary'], 
                 parameters_dict=None, name='circular_boundary', info=None,
                 viol_buffer=1e-3):
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.x_cen = x_cen
        self.y_cen = y_cen
        self.radius = radius
        self.viol_buffer = viol_buffer


    def cal_viol_degree(self, x, y):
        
        x_list = np.asarray(x).ravel()
        y_list = np.asarray(y).ravel()

        dist2cen = np.sqrt((x_list - self.x_cen)**2 + (y_list - self.y_cen)**2)
        viol_degree = np.where(dist2cen - self.viol_buffer > self.radius, 
                               dist2cen - self.radius, 
                               0.0)
        
        if np.isscalar(x):
            return viol_degree[0]
        elif isinstance(x, np.ndarray):
            return viol_degree.reshape(x.shape)
        else:
            return  [viol_degree[i] for i in range(len(x_list))]
    

    def generate_boundary_points(self, num_points=100):
        theta_range = np.linspace(0.0, np.pi*2, num=num_points, endpoint=True)
        x_boundary = self.x_cen + self.radius * np.cos(theta_range)
        y_boundary = self.y_cen + self.radius * np.sin(theta_range)
        return x_boundary, y_boundary
    
    
    def func(self, input_dict):
        x = input_dict['layout_x']
        y = input_dict['layout_y']

        viol_degree = np.sum(self.cal_viol_degree(x, y))

        output_dict = {'viol_boundary': viol_degree}

        return output_dict
    


class MinimalSpacing(Model):
    def __init__(self, minimal_spacing,
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['viol_spacing'], 
                 parameters_dict=None, name='circular_boundary', info=None):
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.minimal_spacing = minimal_spacing

    
    def cal_viol_degree(self, x, y):
        x_list = np.asarray(x).ravel()
        y_list = np.asarray(y).ravel()
        viol_degree = np.zeros_like(x_list) + 0.0

        for i in range(len(x_list)):
            dist2turbines = np.sqrt((x_list[i] - x_list)**2 +
                                    (y_list[i] - y_list)**2)
            
            viol_degree_list = np.where(dist2turbines < self.minimal_spacing,
                         self.minimal_spacing - dist2turbines, 0.0)
            
            viol_degree_list[i] = 0 # distance to itself not considered

            viol_degree[i] = np.sum(viol_degree_list)
        
        if np.isscalar(x):
            return viol_degree[0]
        elif isinstance(x, np.ndarray):
            return viol_degree.reshape(x.shape)
        else:
            return  [viol_degree[i] for i in range(len(x_list))]
        
    
    def func(self, input_dict):
        x = input_dict['layout_x']
        y = input_dict['layout_y']

        viol_degree = np.sum(self.cal_viol_degree(x, y))

        output_dict = {'viol_spacing': viol_degree}

        return output_dict
    


if __name__ == '__main__':

    def square(x):
        y = x**2
        return y
    
    def square_decorated(input_dict):
        x = input_dict['x']
        y = square(x)
        output_dict = {'y': y}
        return output_dict
    
    model_square = Model(func=square_decorated,
                         input_arg_keys=['x'],
                         output_arg_keys=['y'],
                         parameters_dict=None)
    





        
    






# -*- coding: utf-8 -*-
"""
Created on Wed Jan 12 16:02:52 2022

@author: jufen
"""

import numpy as np
import os
import time


class RandomSearch(object):
    """ Random search algorithm for wind farm layout optimization. Changing 
    turbine locations (x_i, y_i) for num_dim=2 or turbine locations and mooring
    line configurations (x_i, y_i, alpha_i) for num_dim=3.

    Referece of the algorithm:
        Feng, J. and Shen, W.Z., 2015. Solving the wind farm layout 
        optimization problem using random search algorithm. Renewable Energy, 
        78, pp.182-192. [https://doi.org/10.1016/j.renene.2015.01.005]
    """
    def __init__(self, fun_obj, 
                 lower_bounds, upper_bounds, fun_constr=None, num_dim=3,
                 max_move_step=None, save_path=None):
        self.fun_obj = fun_obj
        self.fun_constr = fun_constr
        self.lower_bounds = lower_bounds # [x_min, y_min] or [x_min, y_min, 
                                         # alpha_min] for num_dim=2 or 3 
        self.upper_bounds = upper_bounds
        self.num_dim=num_dim
        
        if max_move_step is None:
            max_move_step = np.sqrt((upper_bounds[0] - lower_bounds[0])**2 +
                                    (upper_bounds[1] - lower_bounds[1])**2)
        self.max_move_step = max_move_step
        if save_path is None:
            save_path = './'
        
        self.make_save_path(save_path)
        

    def make_save_path(self, save_path):
        self.save_path = save_path

        if not os.path.isdir(self.save_path):
            os.mkdir(self.save_path)


    def randomly_move_a_turbine(self, x_i, y_i, i_move):
        in_bounds = False
        
        while not in_bounds:
            move_dir = np.random.rand() * 2 * np.pi
            move_step = np.random.rand() * self.max_move_step
            new_x = x_i[i_move] + move_step * np.cos(move_dir)
            new_y = y_i[i_move] + move_step * np.sin(move_dir)
            
            in_bounds = (
                new_x <= self.upper_bounds[0] and new_x >= self.lower_bounds[0]
            and new_y <= self.upper_bounds[1] and new_y >= self.lower_bounds[1])
        x_i_new = x_i + 0
        y_i_new = y_i + 0
        x_i_new[i_move] = new_x
        y_i_new[i_move] = new_y
        return x_i_new,  y_i_new


    def randomly_change_a_configuration(self, alpha_i, i_move):
        right_prob = np.random.rand()
        if right_prob > 0.5:
            alpha_new = alpha_i[i_move] + (
                self.upper_bounds[2] - alpha_i[i_move]) * np.random.rand()
        else:
            alpha_new = alpha_i[i_move] - (
                alpha_i[i_move] - self.lower_bounds[2]) * np.random.rand()
                
        alpha_i_new = alpha_i + 0
        alpha_i_new[i_move] = alpha_new
        return alpha_i_new


    def generate_feasible_new_design(self, x, max_tries=1000,
                                     change_same_wt=True):
        i_try = 0

        if self.num_dim == 3:
            num_wt = len(x)//3
            x_i = x[:num_wt]
            y_i = x[num_wt:num_wt*2]
            alpha_i = x[num_wt*2:num_wt*3]
        
            vio_constr = True
            while i_try < max_tries and vio_constr:
                i_move = np.random.randint(num_wt)
                x_i_new, y_i_new = self.randomly_move_a_turbine(x_i, y_i, i_move)

                if change_same_wt:
                    i_change = i_move
                else:
                    i_change = np.random.randint(num_wt)
                alpha_i_new = self.randomly_change_a_configuration(alpha_i, i_change)
                
                x_new = np.hstack((x_i_new, y_i_new, alpha_i_new))
                if self.fun_constr is None:
                    vio_constr = False
                else:
                    vio_constr = self.fun_constr(x_new) > 0

        elif self.num_dim == 2:
            num_wt = len(x)//2
            x_i = x[:num_wt]
            y_i = x[num_wt:num_wt*2]
        
            vio_constr = True
            while i_try < max_tries and vio_constr:
                i_move = np.random.randint(num_wt)
                x_i_new, y_i_new = self.randomly_move_a_turbine(x_i, y_i, i_move)
                
                x_new = np.hstack((x_i_new, y_i_new))
                if self.fun_constr is None:
                    vio_constr = False
                else:
                    vio_constr = self.fun_constr(x_new) > 0
    
        if vio_constr:
            return x
        else:
            return x_new


    def rs_run(self, x0, num_iter=100, max_move_step=None, change_same_wt=True,
               disp_every_step=True):
        if max_move_step is not None:
            self.max_move_step = max_move_step
        start = time.time()
        x_current = x0 + 0
        i = 0
        obj_current = self.fun_obj(x_current)
        hist = [np.hstack((i, x_current, obj_current))]
    
        while i < num_iter:
            x_new = self.generate_feasible_new_design(x_current,
                                                      change_same_wt)
            obj_new = self.fun_obj(x_new)
            i = i + 1
            
            if obj_new < obj_current:
                x_current = x_new
                obj_current = obj_new
                hist.append(np.hstack((i, x_current, obj_current)))
                np.savetxt(os.path.join(
                    self.save_path, 'hist_rs_current.txt'), hist)
                if not disp_every_step:
                    print(f'{i} / {num_iter}, obj={obj_current};')
            if disp_every_step:        
                print(f'{i} / {num_iter}, obj={obj_current};')

        cpu_time = time.time() - start
        np.savetxt(os.path.join(
            self.save_path, 'hist_rs.txt'), hist)
        np.savetxt(os.path.join(
            self.save_path, 'best_solution_rs.txt'),
            np.hstack((x_current, obj_current)))
        
        with open(os.path.join(self.save_path, 'general_info.txt'), 'w') as f:
            f.write('max_iteration: {0}\n'.format(num_iter)  + 
                    'max_move_step: {0}\n'.format(self.max_move_step) +
                    'cpu_time: {0}\n'.format(cpu_time))

        return x_current, obj_current, hist
        
        
        
        
        


import numpy as np
import os

class PSO(object):
    """
    Particle swarm optimization algorithm for continuous optimization problems.
    Assume constraints are included in the objective function using static 
    penalty terms.

    Reference of the algorithm and default parameters:
        Chowdhury, S., Zhang, J., Messac, A. and Castillo, L., 2012. 
        Unrestricted wind farm layout optimization (UWFLO): Investigating key 
        factors influencing the maximum power generation. Renewable Energy, 
        38(1), pp.16-30. [https://doi.org/10.1016/j.renene.2011.06.045]
    """
    
    def __init__(self, obj_fun, num_design_variables, 
                 lower_bounds, upper_bounds, 
                 pop_size=None, max_gen=10,
                 alpha=0.5, beta_l=1.4, beta_g=1.4,
                 save_path=None):
        self.obj_fun = obj_fun
        self.num_design_variables = num_design_variables
        self.lower_bounds = lower_bounds
        self.upper_bounds = upper_bounds
        if pop_size is None:
            self.pop_size = 5 * self.num_design_variables
        else:
            self.pop_size = pop_size
        self.max_gen = max_gen
        self.i_gen = 0
            
        self.alpha = alpha
        self.beta_l = beta_l
        self.beta_g = beta_g
        self.hist = []
        
        if save_path is None:
            save_path = './'
        self.save_path = save_path
        
        if not os.path.isdir(self.save_path):
            os.mkdir(self.save_path)
        
        
    def initialize_pop(self, feedin_solutions=None):
        """ Initialize the population with random solutions. feedin_solutions 
        can pass in specified initial solutions (as a list of solutions).
        """
        self.current_pop = np.zeros((self.pop_size, self.num_design_variables+1))
        
        for i_variable in range(self.num_design_variables):
            self.current_pop[:, i_variable] = (
                    (self.lower_bounds[i_variable] +
                            np.random.rand(self.pop_size) * 
                            (self.upper_bounds[i_variable] - 
                             self.lower_bounds[i_variable])))
        
        if feedin_solutions is not None:
            for j in range(self.pop_size):
                self.current_pop[j, :-1] = feedin_solutions[j]
                
                
        for k in range(self.pop_size):
            self.current_pop[k, -1] = self.obj_fun(
                    self.current_pop[k, :-1])
        
        self.best_pop = (np.zeros_like(self.current_pop) +
                         self.current_pop)
        self.best_global = self.best_pop[np.argmin(
                self.best_pop[:, -1]), :] 
        self.v_current = np.zeros((self.pop_size,
                                   self.num_design_variables))
        np.savetxt(os.path.join(self.save_path, 'initial_pop.txt'),
                   self.current_pop)
        self.record()
   

    def record(self):
        self.hist.append([np.min(self.best_pop[:, -1]), 
                          np.max(self.best_pop[:, -1]),
                          np.mean(self.best_pop[:, -1]),
                          np.std(self.best_pop[:, -1]),
                          np.min(self.current_pop[:, -1]), 
                          np.max(self.current_pop[:, -1]),
                          np.mean(self.current_pop[:, -1]),
                          np.std(self.current_pop[:, -1]),
                          ])

    
    def save_current_best(self):
        i_best = np.argmin(self.current_pop[:, -1])
        np.savetxt(os.path.join(self.save_path, 'current_best_solution.txt'),
                   self.current_pop[i_best, :])
        np.savetxt(os.path.join(self.save_path, 'current_hist.txt'),
                   self.hist)
        with open(os.path.join(self.save_path, 'current_info.txt'), 'w') as f:
            f.write('gen: {0}\n'.format(self.i_gen)  + 
                    'pop_size: {0}\n'.format(self.pop_size) +
                    'alpha: {0}\n'.format(self.alpha) +
                    'beta_g: {0}\n'.format(self.beta_g) +
                    'beta_l: {0}\n'.format(self.beta_l))

        
    def evolve(self):
        self.v_current = (self.alpha*self.v_current + self.beta_l *
                          np.random.rand() * (self.best_pop[:, :-1] -
                          self.current_pop[:, :-1]) + self.beta_g *           
                          np.random.rand() * 
                          (self.best_global[np.newaxis, :-1] -
                           self.current_pop[:, :-1]))
        self.current_pop[:, :-1] = (self.current_pop[:, :-1] +
                        self.v_current)
        
        for k in range(self.pop_size):
            self.current_pop[k, :-1] = np.where(
                    self.current_pop[k, :-1] < self.lower_bounds,
                    self.lower_bounds, self.current_pop[k, :-1])
            self.current_pop[k, :-1] = np.where(
                    self.current_pop[k, :-1] > self.upper_bounds,
                    self.upper_bounds, self.current_pop[k, :-1])
            
            self.current_pop[k, -1] = self.obj_fun(
                    self.current_pop[k, :-1])
            
            if self.current_pop[k, -1] < self.best_pop[k, -1]:
                self.best_pop[k, :] = self.current_pop[k, :]
        
        self.best_global = self.best_pop[
                np.argmin(self.best_pop[:, -1]), :]
        self.record()
        self.save_current_best()
        self.i_gen += 1
  
    
    def pso_run(self, feedin_solutions=None):
        
        self.initialize_pop(feedin_solutions)
        while self.i_gen < self.max_gen:
            self.evolve()
            print('Gen {0}/{1}, '.format(self.i_gen, self.max_gen) + 
                  'Global best obj = {0:.6f}, '.format(
                          self.best_global[-1]) +
                  'Current best obj = {0:.6f}, '.format(
                          min(self.current_pop[:, -1])) +
                  'Current worst obj = {0:.6f}'.format(
                          max(self.current_pop[:, -1]))
                          )
        
        best_solution = self.best_global[:-1]
        best_obj = self.best_global[-1]

        return best_solution, best_obj, self.hist
        
 

if __name__ == '__main__': 
    
    def obj_fun(x):
        return (x[0]-2)**2 + (x[1]-1)**2
    
    num_design_variable = 2
    lower_bounds = np.array([-5, -5])
    upper_bounds = np.array([5, 5])
    
    PSO_opt = PSO(obj_fun, num_design_variable, 
                       lower_bounds, upper_bounds, max_gen=100)
        
    PSO_opt.pso_run()
    
    


from topdesign import Model



class System(Model):
    def __init__(self, model_dict={}, workflow_graph=[['input', 'output']], 
                 input_arg_keys=['x'], output_arg_keys=['y'], 
                 parameters_dict=None, name='work_flow', info=None):
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)

        self.model_dict = model_dict
        self.workflow_graph = workflow_graph


    



import sys
import numpy as np
from topdesign import System



class WFEvalSystem(System):
    """ 
    Reference of the concept:
        Feng, J. and Sørensen, J.N., 2024, June. Towards a general tool chain 
        integration platform for multi-disciplinary analysis and optimization 
        in wind energy. In Journal of Physics: Conference Series (Vol. 2767, 
        No. 8, p. 082014). IOP Publishing. 
        [https://doi.org/10.1088/1742-6596/2767/8/082014]
    """
    def __init__(self, model_dict={}, workflow_graph=None, 
                 input_arg_keys=['layout_x', 'layout_y'], 
                 output_arg_keys=['obj'],
                 wf_model_type='floris',
                 penalty_weighting=1000.0, 
                 parameters_dict=None, name='work_flow', info=None):
        super().__init__(model_dict=model_dict, workflow_graph=workflow_graph,
                         input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.wf_model_type = wf_model_type
        self.penalty_weighting = penalty_weighting
        self.current_AEP = None


    def evaluate(self, input_dict, recording=True):
        viol_boundary_dict = self.model_dict['boundary'].evaluate(input_dict)
        viol_spacing_dict = self.model_dict['spacing'].evaluate(input_dict)

        viol_dict = viol_boundary_dict | viol_spacing_dict

        viol_tot_degree = np.sum(viol_dict['viol_boundary']) + np.sum(
            viol_dict['viol_spacing'])
    
        recalculated_AEP = False 

        if viol_tot_degree <= 0 or self.current_AEP is None:
            if self.wf_model_type == 'floris':
                AEP_dict = self.model_dict['wf_model_floris'].evaluate(
                    input_dict)
                recalculated_AEP = True
            elif self.wf_model_type == 'pywake':
                AEP_dict = self.model_dict['wf_model_pywake'].evaluate(
                    input_dict)
                recalculated_AEP = True
            else:
                raise NotImplementedError
            
            self.current_AEP = AEP_dict['AEP']
        else:
            AEP_dict = {'AEP': self.current_AEP + 0.0}

        output_dict = {'obj': self.cal_obj(AEP_dict['AEP'], viol_tot_degree),
                       'recalculated_AEP': recalculated_AEP}
        output_dict = output_dict | viol_dict | AEP_dict

        if recording:
             self.record_current_run(input_dict, output_dict)
        
        return output_dict


    def cal_obj(self, AEP, viol_tot_degree):
        return -AEP + self.penalty_weighting * viol_tot_degree

        







import numpy as np
import os
import time


class RealCodedGA(object):
    """ Real-coded genetic algorithm based on ranking selection, 
    direction-based crossover and dynbamic random mutation, as proposed in the 
    following paper is implemented here:
        
        Chuang, Yao-Chen, Chyi-Tsong Chen, and Chyi Hwang. "A simple and 
        efficient real-coded genetic algorithm for constrained optimization." 
        Applied Soft Computing 38 (2016): 87-105.
        
    Constraints can be included in obj_fun using static penalty function as
    described in the above paper. Lower and upper bounds for each design
    variables are required.
    """
    def __init__(self, obj_fun, num_design_variables, 
                 lower_bounds, upper_bounds, pop_size=10, max_gen=10,
                 p_remove=None, p_crossover=0.9, mutation_bound=0.5,
                 save_path=None):
        """ p_crossover controls the probability two selected parents being
        crossovered (equal to 1-r in the paper), and mutation_bound controls
        the largest mutation amplitute with regards to the distance between
        upper and lower bounds for each design variable.
        """
        self.obj_fun = obj_fun
        self.num_design_variables = num_design_variables
        self.lower_bounds = lower_bounds
        self.upper_bounds = upper_bounds
        self.pop_size = pop_size
        self.max_gen = max_gen
        self.i_gen = 0
        
        if p_remove is None:
            self.p_remove = 1/self.num_design_variables
        else:
            self.p_remove = p_remove
            
        self.p_crossover = 0.9
        self.mutation_bound = 0.5
        self.hist = []
        
        if save_path is None:
            save_path = './'
        self.save_path = save_path
        
        if not os.path.isdir(self.save_path):
            os.mkdir(self.save_path)
        
        
    def initialize_pop(self, feedin_solutions=None):
        """ Initialize the population with random solutions. feedin_solutions 
        can pass in specified initial solutions (as a list of solutions).
        """
        self.current_pop = np.zeros(
                (self.pop_size, self.num_design_variables+1))
        
        for i_variable in range(self.num_design_variables):
            self.current_pop[:, i_variable] = (
                    self.lower_bounds[i_variable] +
                            np.random.rand(self.pop_size) * 
                            (self.upper_bounds[i_variable] - 
                             self.lower_bounds[i_variable]))
        
        if feedin_solutions is not None:
            for j in range(self.pop_size):
                self.current_pop[j, :-1] = feedin_solutions[j]
                
        for k in range(self.pop_size):
            self.current_pop[k, -1] = self.obj_fun(
                    self.current_pop[k, :-1])
            
        self.rank_pop()    
        self.record()
        np.savetxt(os.path.join(self.save_path, 'initial_pop.txt'),
                   self.current_pop)
        self.new_pop = np.zeros_like(self.current_pop) + self.current_pop
        

    def record(self):
        self.hist.append([np.min(self.current_pop[:, -1]),
                          np.max(self.current_pop[:, -1]),
                          np.mean(self.current_pop[:, -1]),
                          np.std(self.current_pop[:, -1])])
        

    def rank_pop(self):
        self.current_pop = self.current_pop[
                np.argsort(self.current_pop[:, -1]), :]
        
    
    def rank_selection(self):
        self.rank_pop()
        num_remove = int(self.p_remove * self.pop_size)
        if num_remove == 0:
            num_remove = 1
        self.current_pop[-num_remove:, :] = (
                self.current_pop[:num_remove, :])
        self.rank_pop()
        self.obj_max_current = np.max(self.current_pop[:, -1])
        self.obj_min_current = np.min(self.current_pop[:, -1])
        

    def crossover_and_mutation(self):
        
        M = self.pop_size // 2

        for i in range(M):
            solution_A = self.current_pop[i, :]
            solution_B = self.current_pop[i+M, :]
            
            if (np.random.rand() < self.p_crossover and solution_A[-1] != 
                solution_B[-1]):
                solution_A_new, solution_B_new = self.crossover(
                        solution_A, solution_B)
            else:
                solution_A_new, solution_B_new = self.mutation(
                        solution_A, solution_B)
                
            solution_A_new[-1] = self.obj_fun(solution_A_new[:-1])
            solution_B_new[-1] = self.obj_fun(solution_B_new[:-1])    
            self.new_pop[i, :] = solution_A_new
            self.new_pop[i+M, :] = solution_B_new
                
    
    def replacement(self):
        for i_pop in range(self.pop_size):
            if self.new_pop[i_pop, -1] < self.current_pop[i_pop, -1]:
                self.current_pop[i_pop, :] = self.new_pop[i_pop, :]
        
        
    def crossover(self, solution_A, solution_B):
        sc = np.abs(solution_A[-1] - solution_B[-1])/(self.obj_max_current - 
                   self.obj_min_current)
        
        D = np.where(np.random.rand(self.num_design_variables+1)>=0.5, 
                     solution_A - solution_B, 0)
        
        solution_A_new = solution_A + sc*D
        solution_B_new = solution_B + sc*D
 
        return (self.repair_bounds(solution_A_new), 
                self.repair_bounds(solution_B_new))
    

    def mutation(self, solution_A, solution_B):
        sm = 1-((1-self.i_gen)/self.max_gen)**2
        mutation_sizes = -self.mutation_bound + np.random.rand(
                self.num_design_variables)*2*self.mutation_bound
                
        solution_A_new = np.zeros_like(solution_A)
        solution_B_new = np.zeros_like(solution_B)
        
        solution_A_new[:-1] = solution_A[:-1] + sm*mutation_sizes*(
                self.upper_bounds - self.lower_bounds)
        solution_B_new[:-1] = solution_B[:-1] + sm*mutation_sizes*(
                self.upper_bounds - self.lower_bounds)
        
        return (self.repair_bounds(solution_A_new), 
                self.repair_bounds(solution_B_new))
        
    
    def repair_bounds(self, solution):
        solution[:-1] = np.where(solution[:-1] < self.lower_bounds, 
                self.lower_bounds, solution[:-1])
        solution[:-1] = np.where(solution[:-1] > self.upper_bounds, 
                self.upper_bounds, solution[:-1])
        return solution
    
    
    def save_current_best(self, cpu_time):
        i_best = np.argmin(self.current_pop[:, -1])
        np.savetxt(os.path.join(self.save_path, 'current_best_solution.txt'),
                   self.current_pop[i_best, :])
        np.savetxt(os.path.join(self.save_path,  'current_hist.txt'),
                   self.hist)
        with open(os.path.join(self.save_path, 'current_info.txt'), 'w') as f:
            f.write('gen: {0}\n'.format(self.i_gen)  + 
                    'pop_size: {0}\n'.format(self.pop_size) +
                    'p_crossover: {0}\n'.format(self.p_crossover) +
                    'p_remove: {0}\n'.format(self.p_remove) +
                    'mutation_bound: {0}\n'.format(self.mutation_bound) +
                    'cpu_time: {0}\n'.format(cpu_time) )

        
    def evolve(self):
        self.rank_selection()
        self.crossover_and_mutation()
        self.replacement()
        self.record()
        self.i_gen += 1
    

    def ga_run(self, feedin_solutions=None, max_gen=None):
        start = time.time()
        self.initialize_pop(feedin_solutions)
        if max_gen is None:
            max_gen = self.max_gen
    
        while self.i_gen < max_gen:
            self.evolve()
            cpu_time = time.time() - start
            self.save_current_best(cpu_time)
            print('Gen {0}/{1}, '.format(self.i_gen, max_gen) + 
                  'Current best obj = {0:.6f}, '.format(
                          min(self.current_pop[:, -1])) +
                  'worst obj = {0:.6f}'.format(max(self.current_pop[:, -1])))

        i_best = np.argmin(self.current_pop[:, -1])
        best_solution = self.current_pop[i_best, :-1]
        best_obj = self.current_pop[i_best, -1]
        
        return best_solution, best_obj, self.hist
    


if __name__ == '__main__': 
    
    def obj_fun(x):
        return (x[0]-2)**2 + (x[1]-1)**2
    
    num_design_variable = 2
    lower_bounds = np.array([-5, -5])
    upper_bounds = np.array([5, 5])
    
    RCGA_opt = RealCodedGA(obj_fun, num_design_variable, 
                       lower_bounds, upper_bounds, pop_size=50,
                       max_gen=50)
        
    RCGA_opt.ga_run()
    
    


# -*- coding: utf-8 -*-
"""
First created on in 2021.
Updated in Feb. 2023.
@author: Ju Feng
"""

import numpy as np
from scipy.interpolate import interp1d
from math import gamma


def Sector2Subsector(para, axis=-1, wd_binned=None, interp_method='piecewise'):
    """ Expand para on the wind direction dimension, i.e., increase the nubmer
    of sectors (sectors to subsectors), by interpolating between sectors, using
    specified method.

    Parameters
    ----------
    para : array_like
        Parameter to be expand, it can be sector-wise Weibull A, k, frequency.
    axis : integer
        Denotes which dimension of para corresponds to wind direction.
    wd_binned : array_like
        Wind direction of subsectors to be expanded to.
    inter_method : string
        'piecewise'/'linear'/'spline', based on interp1d in scipy.interpolate,
        'spline' means cubic spline.

    --------------------------------------
    Note: the interpolating method for sector-wise Weibull distributions and
    joint distribution of wind speed and wind direction is referred to the
    following paper:
        Feng, J. and Shen, W.Z., 2015. Modelling wind for wind farm layout
        optimization using joint distribution of wind speed and wind direction.
        Energies, 8(4), pp.3075-3092. [https://doi.org/10.3390/en8043075]
    """
    if para is None:
        return None

    if wd_binned is None:
        wd_binned = np.linspace(0, 360, 360, endpoint=False)
    para = np.array(para)
    num_sector = para.shape[axis]
    wd_sector = np.linspace(0, 360, num_sector, endpoint=False)

    try:
        interp_index = ['piecewise', 'linear', 'spline'].index(interp_method)
        interp_kind = ['nearest', 'linear', 'cubic'][interp_index]
    except ValueError:
        raise ValueError(
            'interp_method={0} not implemeted yet.'.format(interp_method))
    wd_sector_extended = np.hstack((wd_sector, 360.0))
    para_sector_extended = np.concatenate((para, para.take([0], axis=axis)),
                                          axis=axis)
    f_interp = interp1d(wd_sector_extended, para_sector_extended,
                                    kind=interp_kind, axis=axis)
    para_expanded = f_interp(wd_binned % 360)

    return para_expanded



class WindCondition_Weibull(object):
    """ Wind condition class defines a uniform wind resource cefined by
    sector-wise Weibull distributions with wind rose at a reference height.
    
    Can be applied to flat terrain or offshore sites.
    """
    def __init__(self, Weibull_A_l, Weibull_k_l, freq_l, h_ref, alpha_ref=0.17, 
                 ti=0.06, theta_l=None, interp_method='spline'):
        self.Weibull_A_l = Weibull_A_l
        self.Weibull_k_l = Weibull_k_l
        self.freq_l = freq_l
        self.h_ref = h_ref
        self.alpha_ref = alpha_ref
        self.ti = ti
        
        self.num_l = len(self.Weibull_A_l)
        if theta_l is None:
            theta_l = np.linspace(0, 360, self.num_l, endpoint=False)
        self.theta_l = theta_l

        self.theta_l_refined = np.linspace(0, 360, 361, endpoint=True)
        self.Weibull_A_l_refined = Sector2Subsector(
            self.Weibull_A_l, wd_binned=self.theta_l_refined,
            interp_method=interp_method)

        self.Weibull_k_l_refined = Sector2Subsector(
            self.Weibull_k_l, wd_binned=self.theta_l_refined,
            interp_method=interp_method)
            
        self.freq_l_refined = Sector2Subsector(
            self.freq_l, wd_binned=self.theta_l_refined,
            interp_method=interp_method)
        
        d_theta = self.theta_l_refined[1] - self.theta_l_refined[0]
        self.freq_l_refined[:-1] = self.freq_l_refined[:-1] / np.sum(
            self.freq_l_refined[:-1])
        self.freq_l_refined[-1] = self.freq_l_refined[0]
        self.freq_l_refined = self.freq_l_refined/d_theta

        self.interp_Weibull_A = interp1d(
            self.theta_l_refined, self.Weibull_A_l_refined, kind='linear')
        self.interp_Weibull_k = interp1d(
            self.theta_l_refined, self.Weibull_k_l_refined, kind='linear')
        self.interp_freq = interp1d(
            self.theta_l_refined, self.freq_l_refined, kind='linear')


    def get_pdf(self, ws, wd, h=None):
        if h is not None:
            ws = ws * (self.h_ref / h) ** self.alpha_ref

        wd = np.mod(wd, 360)
        Weibull_A = self.interp_Weibull_A(wd)
        Weibull_k = self.interp_Weibull_k(wd)
        freq = self.interp_freq(wd)
        pdf =(Weibull_k / Weibull_A * (ws/Weibull_A)**(Weibull_k - 1) * 
              np.exp(- (ws/Weibull_A)**Weibull_k) * freq)
        return pdf
    
    @classmethod
    def from_time_series(cls, ws_ts, wd_ts, h_ref, num_l=12, alpha_ref=0.17, 
                 ti=0.06, interp_method='linear'):
        
        wd_ts = np.mod(wd_ts, 360)
        theta_l = np.linspace(0, 360, num_l, endpoint=False)
        sector_width = 360.0 / num_l
        Weibull_A_l = np.zeros(num_l)
        Weibull_k_l = np.zeros(num_l)
        freq_l = np.zeros(num_l)

        num_points = len(ws_ts)

        for i in range(num_l):
            theta_min = theta_l[i] - sector_width / 2
            theta_max = theta_l[i] + sector_width / 2

            if num_l == 0:
                index_l = np.where((wd_ts >= 360 + theta_min) | 
                                   (wd_ts < theta_max))[0]
            else:
                index_l = np.where((wd_ts >= theta_min) & 
                                   (wd_ts < theta_max))[0]
                
            ws_l = ws_ts[index_l]
            wd_l = wd_ts[index_l]

            ws_mean_l = np.mean(ws_l)
            ws_std_l = np.std(ws_l)

            Weibull_k_l[i] = (ws_std_l / ws_mean_l) ** (-1.086) 
            Weibull_A_l[i] = ws_mean_l / (gamma(1 + 1 / Weibull_k_l[i]))
            freq_l[i] = len(ws_l) / num_points

        return cls(Weibull_A_l, Weibull_k_l, freq_l, h_ref, alpha_ref, ti,  
                   theta_l, interp_method)    



if __name__ == '__main__':
    from topdesign import plot_2d_contour, set_font

    ##########################################################################
    # Example with the wind resource at Horns Rev 1 site.
    # Data from the following paper:
    #   Feng, J. and Shen, W.Z., 2015. Modelling wind for wind farm layout
    #   optimization using joint distribution of wind speed and wind direction.
    #   Energies, 8(4), pp.3075-3092. [https://doi.org/10.3390/en8043075]
    
    A = [8.89, 9.27, 8.23, 9.78, 11.64, 11.03, 11.50, 11.92, 11.49, 11.08, 
         11.34, 10.76]
    k = [2.09, 2.13, 2.29, 2.30, 2.67, 2.45, 2.51, 2.40, 2.35, 2.27, 2.24, 
         2.19]
    f = [4.82, 4.06, 3.59, 5.27, 9.12, 6.97, 9.17, 11.84, 12.41, 11.34, 11.70, 
         9.69]
    
    f = [x/100.0 for x in f]
    h_ref = 70

    wind_hornsrev = WindCondition_Weibull(A, k, f, h_ref, 
                                          interp_method='spline')

    ws_range = np.linspace(0, 30, 361)
    wd_range = np.arange(361)
    ws, wd = np.meshgrid(ws_range, wd_range,
                        indexing='ij')

    pdf_hornsrev = wind_hornsrev.get_pdf(ws, wd, h=h_ref)

    set_font(21, 'Time New Roman')

    fig, cbar = plot_2d_contour(pdf_hornsrev.T, extent=[0, 360, 0, 30],
                                xlabel=r'$\theta_{wd}$ [$\degree$]',
                                ylabel=r'$u^\infty$ [m/s]',
                                xticks=np.arange(7)*60,
                                cbar_label=r'$pdf$ [-]')

    fig.savefig('WindCondition_hornsrev.png', bbox_inches='tight')



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 24 10:24:15 2024

@author: jufen
"""

##############################################################################
# Make sure you have py_wake and floris (v4.x) installed

import numpy as np
import matplotlib.pyplot as plt
from py_wake.site import UniformWeibullSite
from py_wake.site.shear import PowerShear
from py_wake.wind_turbines import WindTurbine
from py_wake.wind_turbines.power_ct_functions import PowerCtTabular
from py_wake.literature.gaussian_models import Bastankhah_PorteAgel_2014
from floris import WindRose, FlorisModel
from floris.turbine_library import TurbineInterface
import pandas as pd
import time

import os
import shutil

from topdesign.wind_farm import WFModel_Floris, WFModel_PyWake, WFEvalSystem
from topdesign.wind_farm import CircularBoundary, MinimalSpacing
from topdesign.optimizer import RandomSearch
from topdesign import plot_power_ct_curve, \
    save_PRNG_state_in_pickle, set_PRNG_state_in_pickle




def plot_layout(x, y, wt_pywake, aep_per_turbine, aep_tot,
                x_boundary=None, y_boundary=None, dpi=300):
    
    fig, ax = plt.subplots(dpi=dpi)
    ax.axis("equal")
    wt_pywake.plot(x, y, ax=ax)
    c = ax.scatter(x, y, c=aep_per_turbine)
    if x_boundary is not None:
        plt.plot(x_boundary, y_boundary, 'r--')

    plt.colorbar(c, label="AEP [GWh]")
    plt.title(f"Total AEP: {aep_tot:.2f} GWh")
    plt.xlabel("x [m]")
    plt.ylabel("y [m]")
    
    return fig


if __name__ == '__main__':
    ###########################################################################
    # Handle pseudo random number generator state
    
    # set True if want to reproduce and 'PRNG_state.pkl' is there,
    # otherwise set False
    use_save_PRNG_state = True 

    if use_save_PRNG_state:
        set_PRNG_state_in_pickle('PRNG_state.pkl')
    else:
        save_PRNG_state_in_pickle(save_path='./', 
                                  file_name='PRNG_state.pkl')

    ##########################################################################
    # Wind condition based on Horns Rev 1 data
    Weibull_A = np.array([8.89, 9.27, 8.23, 9.78, 11.64, 11.03, 
                        11.50, 11.92, 11.49, 11.08, 11.34, 10.76])
    
    Weibull_k = np.array([2.09, 2.13, 2.29, 2.30, 2.67, 2.45,
                        2.51, 2.40, 2.35, 2.27, 2.24, 2.19])

    freq = np.array([4.82, 4.06, 3.59, 5.27, 9.12, 6.97, 
                    9.17, 11.84, 12.41, 11.34, 11.70, 9.69])/100
    
    ti = 0.1
    alpha = 0.12
    h_ref = 62 

    h_hub = 150.0

    # upscale Weibull_A to the hub height level
    Weibull_A = Weibull_A * (h_hub/h_ref)**alpha

    #specifying the necessary parameters for the UniformWeibullSite object
    site = UniformWeibullSite(p_wd = freq,      # sector frequencies
                            a = Weibull_A,    # Weibull scale parameter
                            k = Weibull_k,    # Weibull shape parameter
                            ti = ti,          # turbulence intensity, optional
                            shear=PowerShear(h_ref=h_hub, alpha=alpha))

    ws = np.arange(30) * 1.0
    wd = np.arange(360) * 1.0

    P_ilk = site.local_wind(x=[0], y=[0], h=h_hub, wd=wd, ws=ws)['P_ilk']

    freq_list = []

    for k in range(len(ws)):
        for l in range(len(wd)):
            freq_list.append([ws[k], wd[l], P_ilk[0, l, k]])

    # make folder for outputs
    if not os.path.isdir('outputs'):
        os.mkdir('outputs')
            
    df = pd.DataFrame(data = freq_list, columns=['ws', 'wd', 'freq_val'])
    df.to_csv('outputs/wind_rose_horns_rev.csv', index=False)

    # Read in the wind rose using the class
    wind_rose = WindRose.read_csv_long(
        "outputs/wind_rose_horns_rev.csv", wd_col="wd", ws_col="ws", 
        freq_col="freq_val", ti_col_or_value=ti)

    # Show the wind rose
    fig_wind = wind_rose.plot(wd_step=15)
    plt.savefig('outputs/wind_rose.png', dpi=300)
    plt.close()


    # ##############################################################################
    # # Wind turbine object of IEA 15MW reference turbine
    wt_floris = TurbineInterface.from_library("internal", "iea_15MW.yaml")
    D = wt_floris.turbine.rotor_diameter

    power_range = wt_floris.power_curve(ws)[1]
    Ct_range = wt_floris.thrust_coefficient_curve(ws)[1]

    wt_pywake = WindTurbine(name='iea-15MW',
                            diameter=D,
                            hub_height=h_hub,
                            powerCtFunction=PowerCtTabular(
                                ws, power_range, 'MW', Ct_range))

    wt_rated_power = 15 # 15 MW

    fig_wt = plot_power_ct_curve(ws, power_range, Ct_range, 
                                 title='IEA 15MW reference turbine')
    plt.savefig('outputs/wind_turbine.png', dpi=300)
    plt.close(fig_wt)


    ###########################################################################
    # Wind farm models with PyWake and Floris
    wf_obj_pywake = Bastankhah_PorteAgel_2014(site, wt_pywake, k=0.0324555)

    parameters_dict = {'wd': wd, 'ws': ws}

    wf_model_pywake = WFModel_PyWake(wf_obj=wf_obj_pywake,
                                    parameters_dict=parameters_dict)

    # Load the FLORIS object
    wf_obj_floris = FlorisModel("inputs/gch.yaml")
    wf_obj_floris.set(wind_data=wind_rose)

    wf_model_floris = WFModel_Floris(wf_obj=wf_obj_floris)

    
    ###########################################################################
    # Initial layout of 9 wts
    layout_x = np.array(
        [0.0, 5, 10, 0.0, 5, 10, 0.0, 5, 10]) * D *1.5
    layout_y = np.array(
            [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 10.0, 10.0, 10.0]) * D *1.5

    input_dict = {'layout_x': layout_x,
                  'layout_y': layout_y}


    ###########################################################################
    # Define constraints
    x_cen = layout_x[4]
    y_cen = layout_y[4]
    radius = np.sqrt(7.5**2 + 7.5**2) * D
    wf_boundary = CircularBoundary(x_cen, y_cen, radius)

    mini_spacing = 4 * D
    wf_spacing = MinimalSpacing(mini_spacing)


    ###########################################################################
    # Define wind farm evaluation system
    wf_model_type = 'pywake'

    wf_eval_system = WFEvalSystem(
        model_dict={'boundary': wf_boundary,
                    'spacing': wf_spacing,
                    'wf_model_floris': wf_model_floris,
                    'wf_model_pywake': wf_model_pywake},
        wf_model_type=wf_model_type)
    
    wf_eval_system_floris = WFEvalSystem(
        model_dict={'boundary': wf_boundary,
                    'spacing': wf_spacing,
                    'wf_model_floris': wf_model_floris,
                    'wf_model_pywake': wf_model_pywake},
        wf_model_type='floris')


    ###########################################################################
    # Define optimization problem

    # function to parse design vector to x and y cooridnates
    def parse_design_vector(x):
        num_wt = len(x) // 2
        x_i = x[:num_wt]
        y_i = x[num_wt:]
        return x_i, y_i
            
    # objective function
    def fun_obj(x):
        x_i, y_i = parse_design_vector(x)
        input_dict = {'layout_x': x_i,
                'layout_y': y_i}
        output_dict = wf_eval_system.evaluate(input_dict)
        return output_dict['obj']
    
    # objective function with floris
    def fun_obj_floris(x):
        x_i, y_i = parse_design_vector(x)
        input_dict = {'layout_x': x_i,
                'layout_y': y_i}
        output_dict = wf_eval_system_floris.evaluate(input_dict)
        return output_dict['obj']

    # lower/upper bounds based on the circular boundary
    lower_bounds = np.array([x_cen - radius, y_cen - radius])
    upper_bounds = np.array([x_cen + radius, y_cen + radius])

    # Define optimization problem
    rs_opt = RandomSearch(fun_obj=fun_obj, lower_bounds=lower_bounds, 
                          upper_bounds=upper_bounds, fun_constr=None,
                          num_dim=2, save_path='outputs')
    
    
    
    #######################################################################
    # run optimization

    num_wt = len(layout_x)
    num_iter = 2000
    max_move_step = 6000.0
    num_runs = 3

    for i_run in range(num_runs):
        tic = time.time()

        save_path = f'outputs/{num_wt}wt_{num_iter}iter_run{i_run}_RS'

        rs_opt.make_save_path(save_path)

        res = rs_opt.rs_run(x0=np.hstack((layout_x, layout_y)), 
                            num_iter=num_iter,
                            max_move_step=max_move_step, disp_every_step=True)
        
        toc = time.time()

        cpu_time_RS = toc-tic

        print('Optimization took: {:.2f}s'.format(cpu_time_RS))

        cf = -res[1]/(num_wt*15*8760/1000)
        print(res[1])
        print(f'capacity factor = {cf*100:.2f} %')
        print(wf_eval_system.model_dict['wf_model_floris'].num_eval)
        print(wf_eval_system.model_dict['wf_model_pywake'].num_eval)
        num_AEP_calculations = wf_eval_system.model_dict[
            'wf_model_pywake'].num_eval + 0
        
        ##############################################################################
        # Post-processing

        x_opt = res[0][:num_wt]
        y_opt = res[0][num_wt:]

        evol_hist = np.vstack(res[-1])

        last_row = evol_hist[-1, :]
        
        if last_row[0] < num_iter-1:
            last_row[0] = num_iter - 1
            evol_hist = np.vstack((evol_hist, last_row))

        fig = plt.figure(dpi=300)
        plt.plot(evol_hist[:, 0], -evol_hist[:, -1])
        plt.xlabel('Iteration [-]')
        plt.ylabel('AEP [GWh]')
        plt.title('Evolutionary history of optimization\n')
        plt.tight_layout()
        plt.savefig(os.path.join(rs_opt.save_path,'evolution_history.png'))
        plt.close(fig)

        # Get boundary points
        x_boundary, y_boundary = wf_boundary.generate_boundary_points()

        # Plot out original layout
        sim_res = wf_obj_pywake(layout_x, layout_y)
        aep = sim_res.aep(normalize_probabilities=False)
        aep_per_turbine = aep.sum(["wd", "ws"])
        aep_init = np.sum(aep_per_turbine)

        fig_init = plot_layout(layout_x, layout_y, wt_pywake, aep_per_turbine,
                               aep_init, x_boundary=x_boundary, 
                               y_boundary=y_boundary)
        plt.savefig(os.path.join(rs_opt.save_path, 'original_layout.png'))
        plt.close(fig_init)

        # Plot out optimized layout
        sim_res = wf_obj_pywake(x_opt, y_opt)
        aep = sim_res.aep(normalize_probabilities=False)
        aep_per_turbine = aep.sum(["wd", "ws"])
        aep_opt = np.sum(aep_per_turbine)

        fig_RS = plot_layout(x_opt, y_opt, wt_pywake, aep_per_turbine,
                             aep_opt, x_boundary=x_boundary, 
                             y_boundary=y_boundary)
        plt.savefig(os.path.join(rs_opt.save_path, 'optimized_layout.png'))
        plt.close(fig_RS)

        # Recaculating with floris model
        evol_hist_recalculated = evol_hist + 0.0

        for i in range(len(evol_hist_recalculated[:, 0])):
            evol_hist_recalculated[i, -1] = fun_obj_floris(
                evol_hist_recalculated[i, 1:-1])

        # Plot out evolutionary histories
        fig = plt.figure(dpi=300)
        plt.plot(evol_hist[:, 0], -evol_hist[:, -1],
                 evol_hist_recalculated[:, 0], -evol_hist_recalculated[:, -1])
        plt.legend(['PyWake', 'FLORIS'])
        plt.xlabel('Iteration [-]')
        plt.ylabel('AEP [GWh]')
        plt.title('Evolutionary history of optimization\n')
        plt.tight_layout()
        plt.savefig(os.path.join(rs_opt.save_path,'evolution_history_both.png'))
        plt.close(fig)


        # Summary text
        full_aep = wt_rated_power * num_wt * 8760 / 1000

        summary_str = []
        summary_str.append('Summary of results\n')

        summary_str.append(f'i_run = {i_run}')
        summary_str.append(f'cpu_time = {cpu_time_RS:.4f} s')
        summary_str.append(f'save_path: {save_path}')

        summary_str.append(f'num_wt = {num_wt}')
        summary_str.append(f'num_iter = {num_iter}')
        summary_str.append(f'num_AEP_calculations = {num_AEP_calculations}')
        summary_str.append(f'max_move_step = {max_move_step}\n')
        
        # %% Calculate ideal AEP (without wake loss)

        xy_single_turbine = np.array([layout_x[0], layout_y[0]])

        summary_str.append('\n------------PyWake------------------')
        aep_single = - fun_obj(xy_single_turbine)
        aep_ideal = aep_single * num_wt 

        wake_loss_init = (1 - aep_init/aep_ideal) * 100
        wake_loss_opt = (1 - aep_opt/aep_ideal) * 100
        imp_pct = (aep_opt/aep_init - 1) * 100

        summary_str.append(f'AEP_gross = {aep_ideal:.4f} GWh')
        summary_str.append(f'AEP_initi = {aep_init:.4f} GWh, CF = {aep_init/full_aep*100:.2f} %')
        summary_str.append(f'AEP_optim = {aep_opt:.4f} GWh, CF = {aep_opt/full_aep*100:.2f} %')
        summary_str.append(f'wake_loss_initi = {wake_loss_init:.4f} %')
        summary_str.append(f'wake_loss_optim = {wake_loss_opt:.4f} %')
        summary_str.append(f'AEP increase percentage = {imp_pct:.4f} %')


        summary_str.append('\n------------Floris------------------')
        aep_single = - fun_obj_floris(xy_single_turbine)
        aep_ideal = aep_single * num_wt 

        aep_init = -evol_hist_recalculated[0, -1]
        aep_opt = -evol_hist_recalculated[-1, -1]

        wake_loss_init = (1 - aep_init/aep_ideal) * 100
        wake_loss_opt = (1 - aep_opt/aep_ideal) * 100
        imp_pct = (aep_opt/aep_init - 1) * 100

        summary_str.append(f'AEP_gross = {aep_ideal:.4f} GWh')
        summary_str.append(f'AEP_initi = {aep_init:.4f} GWh, CF = {aep_init/full_aep*100:.2f} %')
        summary_str.append(f'AEP_optim = {aep_opt:.4f} GWh, CF = {aep_opt/full_aep*100:.2f} %')
        summary_str.append(f'wake_loss_initi = {wake_loss_init:.4f} %')
        summary_str.append(f'wake_loss_optim = {wake_loss_opt:.4f} %')
        summary_str.append(f'AEP increase percentage = {imp_pct:.4f} %')


        print('\n'.join(summary_str))

        with open(os.path.join(rs_opt.save_path, 'results_summary.txt'), 'w') as f:
            f.write('\n'.join(summary_str))
        
        # copy the current script to the result saving path
        shutil.copy(__file__, os.path.join(rs_opt.save_path, 
                                           'main_script_for_this_run.py')) 


###############################################################################
# To run this example, you need to have the following library installed:
# 1. matlabengine (https://pypi.org/project/matlabengine/)
#    [check here for more details: 
#          https://www.mathworks.com/help/matlab/matlab-engine-for-python.html]
#
# 2. juliacall (https://pypi.org/project/juliacall/)
#    [check here for more details: 
#          https://juliapy.github.io/PythonCall.jl/stable/juliacall/]

###############################################################################
# This example solves test problem HS15 from the Hock & Schittkowski collection
#
#   min   100 (x2 - x1^2)^2 + (1 - x1)^2
#   s.t.  x1 x2 >= 1
#         x1 + x2^2 >= 0
#         x1 <= 0.5
#
# assuming the objective function is an executable based on Fortran code,
# the 1st constraint is a Julia function and the 2nd constraint is a matlab 
# function and the 3rd constraint is included in the upper bound of x1.
#
# Ref.: W. Hock and K. Schittkowski, Test Examples for Nonlinear Programming
# Codes.  Lecture Notes in Economics and Mathematical Systems, v. 187,
# Springer-Verlag, New York, 1981, p. 38.



import numpy as np
from topdesign import ModelMatlab, ModelJulia, ModelExecutableStdIO
from topdesign.optimizer import RandomSearch, PSO, RealCodedGA



if __name__ == '__main__':


    ###########################################################################
    # 1. objective function model that wraps the executable obj_func.exe stored
    #    in sub_models folder. The original code is written in Fortran.
    obj_func_model = ModelExecutableStdIO(
        executable='sub_models/obj_func.exe',
        input_arg_keys=['x1', 'x2'], 
        output_arg_keys=['obj'])


    ###########################################################################
    # 2. 1st constraint function model that wraps julia function 
    #    (constraint_func1) in 'sub_models/constraint_func1.jl'.
    constraint_func1_model = ModelJulia(
        Julia_func_name='constraint_func1', 
        Julia_script='sub_models/constraint_func1.jl',
        JuliaMain=None,
        input_arg_keys=['x1', 'x2'], 
        output_arg_keys=['constr1'])


    ###########################################################################
    # 3. 2nd constraint function model that wraps the matlab function 
    #    (constraint_func2) in 'sub_models/constraint_func2.m'.
    constraint_func2_model = ModelMatlab(
        matlab_func_name='constraint_func2', 
        matlab_func_folder='sub_models',
        eng=None,
        input_arg_keys=['x1', 'x2'], 
        output_arg_keys=['constr2'])

    input_dict = {'x1': 4.0, 'x2': 5.0}

    print(obj_func_model.evaluate(input_dict))

    print(constraint_func1_model.evaluate(input_dict))

    print(constraint_func2_model.evaluate(input_dict))


    ###########################################################################
    # 4. Optimization using Random Search, PSO, and Real-coded GA.

    def func_obj(x):
        """ Function to calcuate the objective value."""
        input_dict = {'x1': x[0], 'x2': x[1]}
        return obj_func_model.evaluate(input_dict)['obj']


    def func_constr(x):
        """ Function to calcuate the constraint violation degree. """
        input_dict = {'x1': x[0], 'x2': x[1]}
        constr1 = constraint_func1_model.evaluate(input_dict)['constr1']
        constr2 = constraint_func2_model.evaluate(input_dict)['constr2']
        constr_violation = max(0, -constr1) + max(0, -constr2)
        return constr_violation


    def func_obj_combined(x):
        """ Combined objective function and constraint violation function with
        a penalty method. """
        obj = func_obj(x)
        constr_violation = func_constr(x)
        return obj + 1e3 * constr_violation

    # Define the lower and upper bounds of the design variables.
    lower_bounds = np.array([-5, -5])
    upper_bounds = np.array([0.5, 5])


    ###########################################################################
    # Random search optimization.
    rs_opt = RandomSearch(fun_obj=func_obj,
                          lower_bounds=lower_bounds,
                          upper_bounds=upper_bounds,
                          fun_constr=func_constr,
                          num_dim=2,
                          save_path='./RS_results')

    # Initial guess of the design variables.
    x0 = np.array([-2.0, 1.0])

    res = rs_opt.rs_run(x0=x0, max_move_step=5, num_iter=1000)
    print('Optimization using Random Search is done.')
    print(f'Optimized result: x1={res[0][0]}, x2={res[0][1]}, obj={res[1]}')
    print('-----------------------------------------------------------\n\n')

    ###########################################################################
    # PSO optimization.
    pso_opt = PSO(obj_fun=func_obj_combined,
                  num_design_variables=2,
                  lower_bounds=lower_bounds,
                  upper_bounds=upper_bounds,
                  max_gen=100,
                  save_path='./PSO_results')

    res = pso_opt.pso_run()
    print('Optimization using PSO is done.')
    print(f'Optimized result: x1={res[0][0]}, x2={res[0][1]}, obj={res[1]}')
    print('-----------------------------------------------------------\n\n')

    ###########################################################################
    # Real-coded GA optimization.
    rcga_opt = RealCodedGA(obj_fun=func_obj_combined,
                    num_design_variables=2,
                    lower_bounds=lower_bounds,
                    upper_bounds=upper_bounds,
                    max_gen=100,
                    save_path='./RCGA_results')

    res = rcga_opt.ga_run()
    print('Optimization using RCGA is done.')
    print(f'Optimized result: x1={res[0][0]}, x2={res[0][1]}, obj={res[1]}')
    print('-----------------------------------------------------------\n\n')


# -*- coding: utf-8 -*-
"""
Created on Tue Jan 23 13:28:19 2024

@author: Ju Feng (jufen@dtu.dk)
"""


import numpy as np
import subprocess


class Model(object):
    def __init__(self, func=None, input_arg_keys=['x'], output_arg_keys=['y'], 
                 parameters_dict=None, name='unnamed', info=None):
        if func is not None:
            self.func=func
        self.input_arg_keys = input_arg_keys
        self.output_arg_keys = output_arg_keys
        self.parameters_dict = parameters_dict
        self.name = name
        self.info = info

        self.num_args_input = len(self.input_arg_keys)
        self.num_args_output = len(self.output_arg_keys)

        self.history_records = []
        self.num_eval = 0


    def evaluate(self, input_dict, recording=True):
        output_dict = self.func(input_dict)

        if recording:
            self.record_current_run(input_dict, output_dict)
        return output_dict
    

    def record_current_run(self, input_dict, ouput_dict):
        self.history_records.append([input_dict, ouput_dict])
        self.num_eval = self.num_eval + 1

    
    def clean_history_records(self):
        self.history_records = []
        self.num_eval = 0
        
    
    
class Adapter(Model):
    def __init__(self, func=None, input_arg_keys=['x'], output_arg_keys=None,
                 parameters_dict=None, name='adapter_unnamed', info=None):
        
        if func is None:
            def func(input_dict):
                output_dict = input_dict.copy()
                return output_dict
        
        if output_arg_keys is None:
            output_arg_keys = input_arg_keys.copy()
        super().__init__(func, input_arg_keys, output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        
        

class ModelMatlab(Model):
    def __init__(self, matlab_func_name, matlab_func_folder=None, eng=None, 
                 input_arg_keys=['x'], output_arg_keys=['y'],
                 parameters_dict=None, name='matlab_model_unnamed', info=None):
        
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.matlab_func_name = matlab_func_name
        self.matlab_func_folder = matlab_func_folder

        if eng is None:
            try:
                import matlab.engine as engine
            except ImportError:
                print('Warning: Matlab engine is not imported!')
                print('ModelMatlab will not be able to use.')
                
            eng = engine.start_matlab()
        self.eng = eng

        if self.matlab_func_folder is not None:
            s = eng.genpath(self.matlab_func_folder)
            eng.addpath(s, nargout=0)

    
    def func(self, input_dict):
        args2matlab = []

        for key in self.input_arg_keys:
            if key not in input_dict.keys():
                raise ValueError(f'Error: {key} is in the defined ' +
                                 'input_arg_keys, but not provided!')
            args2matlab.append(input_dict[key])
        
        res = eval(f'self.eng.{self.matlab_func_name}' + '(*tuple(args2matlab), ' +
                   'nargout=self.num_args_output)')
        
        output_dict = {}
        
        if self.num_args_output == 1:
            # output from matlab in vector will be in 'matlab.double' format
            # it needs to be converted to numpy array in Python.
            if not (type(res) in [int, float]): 
                res = np.array(res)
                
            output_dict[self.output_arg_keys[0]] = res
            
        else:
            for i in range(self.num_args_output):
                res_i = res[i]
                
                if not (type(res_i) in [int, float]): 
                    res_i = np.array(res[i])
                
                output_dict[self.output_arg_keys[i]] = res_i

        return output_dict
    


class ModelJulia(Model):
    def __init__(self, Julia_func_name, Julia_script, JuliaMain=None,
                 input_arg_keys=['x'], output_arg_keys=['y'],
                 parameters_dict=None, name='Julia_model_unnamed', info=None):
        
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        self.Julia_func_name = Julia_func_name
        self.Julia_script = Julia_script

        if JuliaMain is None:
            try:
                from juliacall import Main as JuliaMain
            except ImportError:
                print('Warning: juliacall is not imported!')
                print('ModelJulia will not be able to use.')
        self.JuliaMain = JuliaMain
        # Include the Julia script so that the function can be found
        self.JuliaMain.include(self.Julia_script)


    def func(self, input_dict):
        args2julia = []

        for key in self.input_arg_keys:
            if key not in input_dict.keys():
                raise ValueError(f'Error: {key} is in the defined ' +
                                 'input_arg_keys, but not provided!')
            args2julia.append(input_dict[key])
        
        res = eval(f'self.JuliaMain.{self.Julia_func_name}' + 
                   '(*tuple(args2julia))')

        output_dict = {}
        
        if self.num_args_output == 1:
            # output from Julia in vector will be in 'juliacall.VectorValue' 
            #format, it needs to be converted to numpy array in Python.
            if not (type(res) in [int, float]): 
                res = np.array(res)
                
            output_dict[self.output_arg_keys[0]] = res
            
        else:
            for i in range(self.num_args_output):
                res_i = res[i]
                
                if not (type(res_i) in [int, float]): 
                    res_i = np.array(res[i])
                
                output_dict[self.output_arg_keys[i]] = res_i

        return output_dict
    
    
    
class ModelExecutableStdIO(Model):
    """ This class wraps an executable that take input arguments from terminal
    (stdin) and return output arguments to terminal (stdout).

    In the current implementation, the input/output arguments are assumed as 
    float numbers. 
    
    The executable is compiled from code written in a compiled language (such 
    as Fortran, C, C++, etc.). 
    """
    def __init__(self, executable, 
                 input_arg_keys=['x'], output_arg_keys=['y'],
                 parameters_dict=None, name='executable_model_unnamed', 
                 info=None):
        
        super().__init__(func=None, input_arg_keys=input_arg_keys, 
                         output_arg_keys=output_arg_keys, 
                         parameters_dict=parameters_dict,
                         name=name, info=info)
        
        self.executable = executable

    
    def func(self, input_dict):
        args2exe = []

        for key in self.input_arg_keys:
            if key not in input_dict.keys():
                raise ValueError(f'Error: {key} is in the defined ' +
                                 'input_arg_keys, but not provided!')
            args2exe.append(input_dict[key])
        
        # Call the executable
        cmd = [self.executable] + [str(arg) for arg in args2exe]
        res = subprocess.run(cmd, stdout=subprocess.PIPE)
        res = res.stdout.decode('utf-8').split()
        res = [float(r) for r in res]

        output_dict = {}
        
        if self.num_args_output == 1:
            output_dict[self.output_arg_keys[0]] = res[0]
        else:
            for i in range(self.num_args_output):
                output_dict[self.output_arg_keys[i]] = res[i]

        return output_dict
        
        
    
if __name__ == '__main__':

    def square(x):
        y = x**2
        return y
    
    def square_decorated(input_dict):
        x = input_dict['x']
        y = square(x)
        output_dict = {'y': y}
        return output_dict
    
    model_square = Model(func=square_decorated,
                         input_arg_keys=['x'],
                         output_arg_keys=['y'],
                         parameters_dict=None)
    





        
    






# -*- coding: utf-8 -*-
"""
Created on Fri Sep 24 12:34:41 2021

@author: jufen
"""


import numpy as np
import os
import pickle
import matplotlib.pyplot as plt
from mpl_toolkits import axes_grid1
import matplotlib as mpl


def rotate_coordinates(x, y, theta):
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    x_rot = x*cos_theta + y*sin_theta
    y_rot = -x*sin_theta + y*cos_theta
    return x_rot, y_rot


def set_font(size=14, font_type='Times New Roman'):

    mpl.style.use('classic')
    # Latex font
    # Say, "the default sans-serif font is COMIC SANS"
    plt.rcParams['font.sans-serif'] = "Times New Roman"
    # Then, "ALWAYS use sans-serif fonts"
    plt.rcParams['font.family'] = "sans-serif"
    #plt.rcParams['font.family'] = font_type # #'STIXGeneral'
    SMALL_SIZE = size
    MEDIUM_SIZE = size + 2
    BIGGER_SIZE = size + 2

    plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
    plt.rc('axes', titlesize=BIGGER_SIZE)    # fontsize of the axes title
    plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
    plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title



def add_colorbar(im, aspect=15, pad_fraction=1.0, **kwargs):
    """Add a vertical color bar to an image plot."""
    divider = axes_grid1.make_axes_locatable(im.axes)
    width = axes_grid1.axes_size.AxesY(im.axes, aspect=1./aspect)
    pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
    current_ax = plt.gca()
    cax = divider.append_axes("right", size=width, pad=pad)
    plt.sca(current_ax)
    return im.axes.figure.colorbar(im, cax=cax, **kwargs)



def plot_2d_contour(array, extent=None, cmap='coolwarm', xlabel=None,
                    ylabel=None, cbar_label=None, title=None,
                    xticks=None, yticks=None,
                    ratio=1.0, cbar_shrink=1, cbar_aspect=20):

    fig, ax = plt.subplots(1, 1)
    im = ax.imshow(np.flipud(array.T), cmap=cmap, extent=extent)

    cbar = None

    xleft, xright = ax.get_xlim()
    ybottom, ytop = ax.get_ylim()
    ax.set_aspect(abs((xright-xleft)/(ybottom-ytop))*ratio)

    cbar = fig.colorbar(im, shrink=cbar_shrink, aspect=cbar_aspect)
    cbar.set_label(cbar_label)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xticks(xticks)
    plt.yticks(yticks)
    plt.title(title)
    plt.tight_layout()

    return fig, cbar



def plot_power_ct_curve(ws, power_range, Ct_range, title=None):
    fig, ax1 = plt.subplots()
    ax1.plot(ws, power_range, 'b-')
    ax1.set_xlabel('Wind speed [m/s]')
    ax1.set_ylabel('Power [MW]', color='b')
    for tl in ax1.get_yticklabels():
        tl.set_color('b')
    plt.ylim(0, max(power_range)*1.1)
    
    ax2 = ax1.twinx()
    ax2.plot(ws, Ct_range, 'r-')
    ax2.set_ylabel('Ct [-]', color='r')
    for tl in ax2.get_yticklabels():
        tl.set_color('r')
    plt.ylim(0, 1)

    if title is not None:
        plt.title(title)
    
    plt.tight_layout()

    return fig



def plot_wind_rose(wind_speed, wind_direction, num_bins=12, speed_bins=None,
                   colormap='viridis', title='Wind rose', 
                   legend_label='Wind Speed [m/s]'):
    """
    Plots a wind rose with different wind speed ranges in different colors.

    Parameters:
        wind_speed (array-like): Array of wind speeds.
        wind_direction (array-like): Array of wind directions in degrees.
        num_bins (int): Number of bins for wind direction.
        speed_bins (array-like): Array of wind speed bin edges.
        colormap (string): colormap used in the wind speed legend, like 
            'viridis', 'coolwarm', etc.
        title (string): title of the figure, default: 'wind rose'.
        legend_label (string): title of the wind speed legend.

    Returns:
        fig
    """
    if speed_bins is None:
        speed_bins = [0, 5, 10, 15, 20, 25, np.inf]  # Default speed bins

    # Convert wind direction to radians
    wind_direction_rad = np.deg2rad(wind_direction)

    # Create bins for wind direction
    direction_bins = np.linspace(0, 2 * np.pi, num_bins + 1) - (np.pi/num_bins)
    direction_bin_centers = (direction_bins[:-1] + direction_bins[1:]) / 2

    # Initialize histogram
    hist = np.zeros((len(speed_bins) - 1, num_bins))

    # Populate histogram
    for i in range(len(speed_bins) - 1):
        mask = (wind_speed >= speed_bins[i]) & (wind_speed < speed_bins[i + 1])
        hist[i], _ = np.histogram(wind_direction_rad[mask], 
                                  bins=direction_bins)

    # Plot wind rose
    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    cm = eval(f'plt.cm.{colormap}')
    colors = cm(np.linspace(0, 1, len(speed_bins) - 1))

    for i in range(len(speed_bins) - 1):
        if speed_bins[i+1] is np.inf:
            label=f'>={speed_bins[i]} m/s'
        else:
            label=f'{speed_bins[i]}-{speed_bins[i+1]} m/s'

        bars = ax.bar(direction_bin_centers, hist[i], 
                      width=(2 * np.pi / num_bins), 
                      bottom=np.sum(hist[:i], axis=0), color=colors[i], 
                      edgecolor='black', label=label)

    # Add labels and title
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_title(title)
    ax.legend(title=legend_label, bbox_to_anchor=(0.92, 1.07), 
              loc='upper left')
    plt.tight_layout()

    return fig



def save_PRNG_state_in_pickle(save_path='.', 
                              file_name='PRNG_state.pkl'):
    state = np.random.get_state()

    state_dict = {
        'algorithm': state[0],
        'keys': state[1],
        'pos': state[2],
        'has_gauss': state[3],
        'cached_gaussian': state[4]}

    with open(os.path.join(save_path, file_name), 
              'wb') as f: 
        pickle.dump(state_dict, f, protocol=pickle.HIGHEST_PROTOCOL)

    return state_dict



def set_PRNG_state_in_pickle(pickle_file='PRNG_state.pkl'):
    with open(pickle_file, 'rb') as f:
        state_dict = pickle.load(f)
    
    state = (state_dict['algorithm'],
             state_dict['keys'],
             state_dict['pos'],
             state_dict['has_gauss'],
             state_dict['cached_gaussian'])
    
    np.random.set_state(state)

    return state



def closestDistanceBetweenLines(
    a0,a1,b0,b1,
    clampAll=True,clampA0=True,clampA1=True,clampB0=True,clampB1=True):

    """ Given two lines defined by numpy.array pairs (a0,a1,b0,b1)
        Return the closest points on each segment and their distance
        
        source: https://stackoverflow.com/questions/2824478/shortest-distance-between-two-line-segments
    """

    # If clampAll=True, set all clamps to True
    if clampAll:
        clampA0=True
        clampA1=True
        clampB0=True
        clampB1=True


    # Calculate denomitator
    A = a1 - a0
    B = b1 - b0
    magA = np.linalg.norm(A)
    magB = np.linalg.norm(B)
    
    _A = A / magA
    _B = B / magB
    
    cross = np.cross(_A, _B);
    denom = np.linalg.norm(cross)**2
    
    
    # If lines are parallel (denom=0) test if lines overlap.
    # If they don't overlap then there is a closest point solution.
    # If they do overlap, there are infinite closest positions, but there is a closest distance
    if not denom:
        d0 = np.dot(_A,(b0-a0))
        
        # Overlap only possible with clamping
        if clampA0 or clampA1 or clampB0 or clampB1:
            d1 = np.dot(_A,(b1-a0))
            
            # Is segment B before A?
            if d0 <= 0 >= d1:
                if clampA0 and clampB1:
                    if np.absolute(d0) < np.absolute(d1):
                        return a0,b0,np.linalg.norm(a0-b0)
                    return a0,b1,np.linalg.norm(a0-b1)
                
                
            # Is segment B after A?
            elif d0 >= magA <= d1:
                if clampA1 and clampB0:
                    if np.absolute(d0) < np.absolute(d1):
                        return a1,b0,np.linalg.norm(a1-b0)
                    return a1,b1,np.linalg.norm(a1-b1)
                
                
        # Segments overlap, return distance between parallel segments
        return None,None,np.linalg.norm(((d0*_A)+a0)-b0)
        
    
    
    # Lines criss-cross: Calculate the projected closest points
    t = (b0 - a0);
    detA = np.linalg.det([t, _B, cross])
    detB = np.linalg.det([t, _A, cross])

    t0 = detA/denom;
    t1 = detB/denom;

    pA = a0 + (_A * t0) # Projected closest point on segment A
    pB = b0 + (_B * t1) # Projected closest point on segment B


    # Clamp projections
    if clampA0 or clampA1 or clampB0 or clampB1:
        if clampA0 and t0 < 0:
            pA = a0
        elif clampA1 and t0 > magA:
            pA = a1
        
        if clampB0 and t1 < 0:
            pB = b0
        elif clampB1 and t1 > magB:
            pB = b1
            
        # Clamp projection A
        if (clampA0 and t0 < 0) or (clampA1 and t0 > magA):
            dot = np.dot(_B,(pA-b0))
            if clampB0 and dot < 0:
                dot = 0
            elif clampB1 and dot > magB:
                dot = magB
            pB = b0 + (_B * dot)
    
        # Clamp projection B
        if (clampB0 and t1 < 0) or (clampB1 and t1 > magB):
            dot = np.dot(_A,(pB-a0))
            if clampA0 and dot < 0:
                dot = 0
            elif clampA1 and dot > magA:
                dot = magA
            pA = a0 + (_A * dot)

    
    return pA,pB,np.linalg.norm(pA-pB)
